
#include <assert.h>
#include <errno.h>

#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/param.h>
#include <sys/wait.h>
#include "file.h"
#include "ee.h"
#include "edit.h"
#include "cli.h"
#include "devurandom.h"
#include "customization.h"
#include "cli_nicks.h"
#include "cli_keys.h"
#include "cli_data.h"
#include "cli_osx.h"

extern key_store_t *keys;

static edit_t eb;
static char  *local_argv[MAX_LOCAL_ARGV + 1];
static int    local_argc;
static pipe_t local_cmd;


int cli_exec_redirect(stage_t *stage, int fd_in, int fd_out, int fd_close)
{
  int pid;

  pid = fork();

  if (0 > pid) {
    fprintf(stderr, "Can't fork: %s\n", strerror(errno));
    return -1;
  }
  if (0 == pid) {
    /*
     * Redirect stdin and stdout.
     */
    if (STDIN_FILENO != fd_in) {
      dup2(fd_in,  STDIN_FILENO);
      close(fd_in);
    }
    if (STDOUT_FILENO != fd_out) {
      dup2(fd_out, STDOUT_FILENO);
      close(fd_out);
    }
    fcntl(STDIN_FILENO, FD_CLOEXEC);
    fcntl(STDOUT_FILENO, FD_CLOEXEC);

    /* Close the dangling input fd.
     */
    if (0 <= fd_close) {
      close(fd_close);
    }
    if (cli_exec != stage->func) {
      exit(stage->func(stage->argc, stage->argv));
    }
    /*
     * We really need to get rid of this mapping if we spawn external
     * commands.  This is where all the keys are and we don't want this to
     * be passed to untrusted code.
     */
    munmap(keys, sizeof(key_store_t));

    /* Off we go...
     */
    execvp(stage->argv[0], stage->argv);
    
    /* If things go well then we never reach this point.
     */
    fprintf(stderr, "Can't run %s: %s (%d)\n", stage->argv[0], strerror(errno), errno);
    exit(-1);
  }
  stage->pid = pid;
  return 0;
}


/* Pre: the argv array is NULL-terminated, that is, argv[argc] == NULL.
 */
int cli_exec(int argc, char **argv)
{
  stage_t stage;
  int status;

  stage.argc = argc;
  stage.argv = argv;
  stage.func = cli_exec;

  if (0 > cli_exec_redirect(&stage, STDIN_FILENO, STDOUT_FILENO, -1)) {
    return -1;
  }
  if (0 > waitpid(stage.pid, &status, 0)) {
    return -1;
  }
  if (!WIFEXITED(status)) {
    return -1;
  }
  return WEXITSTATUS(status);
}


int cli_pipe(pipe_t *cmd)
{
  int fd_pair[2];
  int fd_last;
  int status;
  int err;
  int i;
  int j;

  /* Make sure no pending output gets in the way.
   */
  fflush(NULL);

  /* The last stage outputs to STDOUT.  If we want to, we can as well
   * redirect output to a file or to a socket.
   */
  fd_last = dup(STDOUT_FILENO);

  /* We go through all stages of the pipeline from the end to the front and
   * spawn a process.  Each process gets fd_last as the STDOUT and we give
   * it a STDIN that is connected to a file descriptor that becomes the new
   * fd_last.
   */
  for (i = cmd->max - 1; i >= 0; i--) {
    if (0 > pipe(fd_pair)) {
      break;
    }
    err = cli_exec_redirect(&cmd->stage[i], fd_pair[0], fd_last, fd_pair[1]);

    (void)close(fd_pair[0]);
    (void)close(fd_last);

    if (0 > err) {
      break;
    }
    fd_last = fd_pair[1];
  }
  (void)close(fd_last);

  if (0 <= i) {
    for (j=i+1; j < cmd->max; j++) {
      (void)kill(cmd->stage[j].pid, SIGPIPE);
    }
  }
  /* Wait for all childs to terminate.
   */
  status = 0;
  err = 1;

  for (j=i+1; j < cmd->max; j++) {
    waitpid(cmd->stage[j].pid, &status, 0);

    if (!WIFEXITED(status)) {
      err = -1;
    }
    if (WIFSIGNALED(status)) {
      err = -1;
    }
  }
  if (0 > WEXITSTATUS(status)) {
    err = -1;
  }
  return err;
}


int cli_chdir(int argc, char **argv)
{
  char *path;

  if (2 < argc) {
    fprintf(stderr, "Usage: %s <dir>\n", argv[0]);
    return 0;
  }
  path = (1 == argc) ? getenv("HOME") : argv[1];

  if (chdir(path) < 0) {
    fprintf(stderr, "Can't change to %s: %s\n", path, strerror(errno));
    return -1;
  }
  return 0;
}


int cli_checkpw(int verbose)
{
  uint8_t seed[SEED_LENGTH];
  uint8_t hash[LEN_HASH];
  size_t i;
  char pw[PW_LENGTH];

  /* If a password has been set then we check it and the error counter.
   */
  if (0 == keys->flags) {
    if (0 != verbose) {
      fprintf(stderr, "%s", EES_PWD);
    }
    return 0;
  }
  while (3 > keys->errors) {
    printf("Password entry %d/3 (max %d chars): ", keys->errors+1, PW_LENGTH);
    i = edit_passwd(pw, 64);
    
    memmove(seed, keys->seed, SEED_LENGTH);
    memmove(seed, pw, i);
    crypto_hash(hash, seed, SEED_LENGTH);

    /* Remove sensitive data from memory.
     */
    memset(pw, 0, sizeof(pw));
    memset(&hash[2], 0, sizeof(hash)-2);

    if ((hash[0] == keys->pw[0]) && (hash[1] == keys->pw[1])) {
      keys->errors = 0;
      return 0;
    }
    keys->errors++;
  }
  printf("\nSorry, you failed to enter your password too often.\n");
  printf("Please ask a wizard to reset it.\n");
  return -1;
}


int cli_passwd(int argc, char **argv)
{
  uint8_t seed[SEED_LENGTH];
  uint8_t hash[LEN_HASH];
  char pw1[PW_LENGTH];
  char pw2[PW_LENGTH];
  size_t i;
  size_t j;
  int r;

  if (2 != argc) {
    fprintf(stderr, "Usage: %s (set|clear)\n", argv[0]);
    return 0;
  }
  if (strncmp(argv[1], "clear\0", 6) == 0) {
    keys->pw[0]  = 0;
    keys->pw[1]  = 0;
    keys->flags  = 0;
    keys->errors = 0;
    printf("Password cleared.\n");
    return 1;
  }
  if (strncmp(argv[1], "set\0", 4) != 0) {
    fprintf(stderr, "Usage: %s (set|clear)\n", argv[0]);
    return 0;
  }
  printf("Please enter a password (max. %d chars): ", PW_LENGTH);
  i = edit_passwd(pw1, PW_LENGTH);

  if (0 == i) {
    printf("Aborted, no password set.\n");
    return 0;
  }
  printf("Please re-enter your password: ");
  j = edit_passwd(pw2, PW_LENGTH);
  r = 0;
  
  if ((i != j) || strncmp(pw1, pw2, i) != 0) {
    fprintf(stderr, "Password mismatch, no password set.\n");
    goto fail;
  }
  randombytes(keys->seed, SEED_LENGTH);
  memmove(seed, keys->seed, SEED_LENGTH);
  memmove(seed, pw1, i);
  crypto_hash(hash, seed, SEED_LENGTH);

  keys->pw[0]  = hash[0];
  keys->pw[1]  = hash[1];
  keys->flags  = 1;
  keys->errors = 0;
  r = 1;

  /* Remove sensitive data from memory
   */
 fail:
  memset(hash, 0, sizeof(hash));
  memset(pw1, 0, sizeof(pw1));
  memset(pw2, 0, sizeof(pw2));
  return r;
}


int cli_editor(int argc, char **argv)
{
  char *editor_argv[3];

  if ((1 == argc) || (3 == argc)) {
    fprintf(stderr, "Usage: %s <nick 1> .. <nick n> <file in> <file out>\n", argv[0]);
    fprintf(stderr, "Usage: %s <file>\n", argv[0]);
    return 0;
  }
  editor_argv[0] = argv[0];

  if (2 == argc) {
    editor_argv[1] = argv[1];
  }
  else {
    editor_argv[1] = argv[argc-2];
  }
  if (cli_exec(2, editor_argv) < 0) {
    fprintf(stderr, "Error running %s: %s\n", argv[0], strerror(errno));
    return -1;
  }
  if (2 != argc) {
    if (cli_enc(argc, argv) < 0) {
      return -1;
    }
  }
  return 1;
}


cmd_t commands[] = {
  {0, "cd",      "change directory", &cli_chdir},

  {1, "show",    "show all known nicks", &cli_show},
  {1, "add",     "add a nick (nicks \"private\" and \"passwd\" are special)", &cli_add},
  {1, "delete",  "delete a nick", &cli_delete},
  {1, "rename",  "rename a nick or change tag string", &cli_rename},
  {1, "clone",   "create a new nick with the key of another", &cli_clone},
  {1, "info",    "show information on a nick", &cli_info},

  {2, "enc",     "encrypt a file for some nick", &cli_enc},
  {2, "dec",     "decrypt a file", &cli_dec},
  {2, "export",  "export key to be sent to a nick", &cli_export},
  {2, "import",  "import a key received from a nick", &cli_import},

  {3, "check",   "compare keys over the phone", &cli_check},
  {3, "uncheck", "mark a nick's key as in need of a comparison", &cli_uncheck},
  {3, "transfer","transfer authority from one nick to a new one", &cli_transfer},
  {3, "confirm", "verify a transfer of authority to a new nick", &cli_confirm},

  {4, "backup",  "save an encrypted copy of your key store", &cli_backup},
  {4, "restore", "merge a key store backup into the current one", &cli_restore},

  {5, "passwd",  "set up a password for your keystore", &cli_passwd},
  {5, "quit",    "quit to shell", NULL},
  {5, "exit",    "exit to shell", NULL},
  {5, "help",    "this list", &cli_help},

#ifdef HAVE_OSX
  {6, "attach",  "compose mail with attachment", &cli_attach},
  {6, "fetch",   "save mail attachment", &cli_fetch},
  {6, "view",    "save mail attachment and view it with less", &cli_view},
#endif

  {0, NULL, NULL, NULL}
};


int cli_help(int argc, char **argv)
{
  const char *section[] = {
    "internal",
    "managing contacts",
    "encrypting, decrypting, exchanging keys",
    "checking keys and forward secrecy",
    "backing up and restoring keystores",
    "miscellaneous commands",
#ifdef HAVE_OSX
    "built-in: commands specific to OS X",
#endif
    NULL
  };
  int i;
  int j;

  (void)argc; (void)argv;

  for (j=1; NULL != section[j]; j++) {
    printf(EES_SEP, section[j]);
    
    for (i=0; NULL != commands[i].name; i++) {
      if (j == commands[i].type) {
	printf("%-8s : %s\n", commands[i].name, commands[i].info);
      }
    }
  }
  printf("\n");
  return 1;
}


int cli_is_pipe(char *arg)
{
  return (0 == strncmp(arg, "|\0", 2));
}


int cli_parse()
{
  int j;
  int i;
  int k;
  int u;

  for (i=0, j=0; (i < MAX_LOCAL_PIPE) && (j < local_argc); i++) {
    local_cmd.stage[i].argv = &local_argv[j];

    /* It does, so let's make another stage.
     */
    for (k=j; k < local_argc; k++) {
      if (cli_is_pipe(local_argv[k])) {
	local_argv[k] = NULL;
	break;
      }
    }
    if (k == j) {
      fprintf(stderr, "Invalid input, pipe has an empty stage.\n");
      return -1;
    }
    /* Check whether the command exists.
     */
    for (u=0; NULL != commands[u].name; u++) {
      if (strcmp(commands[u].name, local_argv[j]) == 0) {
	break;
      }
    }
    if (NULL == commands[u].name) {
      local_cmd.stage[i].func = cli_exec;
    }
    else {
      local_cmd.stage[i].func = commands[u].func;

      /* Is this an exit command?
       */
      if (NULL == commands[u].func) {
	return 0;
      }
    }
    /* Everything OK so far, let's move to the next stage.
     */
    local_cmd.stage[i].argc = k - j;
    j = k + 1;
  }
  if (j != (local_argc + 1)) {
    fprintf(stderr, "Invalid input, pipe hast too many stages.\n");
    return -1;
  }
  local_cmd.max = i;
  return 1;
}


int cli_main(void)
{
  int n;

  edit_init(&eb);

  for (;;) {
    edit_reset(&eb);
    edit_line(&eb, EES_PROMPT);

    if (edit_parse(eb.buf, &local_argc, local_argv, MAX_LOCAL_ARGV) < 0) {
      printf("Bad input.\n");
      continue;
    }
    if (0 == local_argc) {
      continue;
    }
    local_argv[local_argc] = NULL;

    n = cli_parse();

    if (0 > n) {
      continue;
    }
    if (0 == n) {
      break;
    }
    if (1 == local_cmd.max) {
      (void)local_cmd.stage[0].func(local_argc, local_argv);
      continue;
    }
    cli_pipe(&local_cmd);
  }
  return 0;
}

