
#include <assert.h>
#include <errno.h>

#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/param.h>
#include <sys/wait.h>
#include "file.h"
#include "ee.h"
#include "viz.h"
#include "encode.h"
#include "edit.h"
#include "keystore.h"
#include "cli.h"
#include "devurandom.h"
#include "tweetnacl.h"
#include "zeit.h"
#include "customization.h"
#include "cli_data.h"

extern key_store_t *keys;


int cli_backup(int argc, char **argv)
{
  uint8_t *buf;
  size_t clen;
  size_t len;
  size_t max;
  size_t h;
  size_t b;
  size_t d;
  int err;

  if ((3 > argc) || (MAX_LOCAL_ARGV < argc)) {
    fprintf(stderr, "Usage: %s <nick 1> .. <nick n> <file out>\n", argv[0]);
    return 0;
  }
  len = sizeof(key_store_t);
  max = argc - 2;
  err = create_efile_helper(len, max, &clen, &h, &b, &d);
  buf = (uint8_t *)malloc(clen);

  if (NULL == buf) {
    fprintf(stderr, "Can't allocate memory for file.\n");
    return -1;
  }
  /* Copy the entire keystore into the allocated memory buffer.
   */
  memcpy(&buf[d], keys, len);

  /* Technically, no error can occur here but, hey, the world is full of
   * weird things...
   */
  if(create_efile(buf, clen, &argv[1], max, h, b) < 0) {
    goto fail;
  }
  err = file_savesync(argv[argc-1], buf, clen);

  if (0 > err) {
    fprintf(stderr, "Error saving file: %s\n", strerror(errno));
    goto fail;
  }
  err = 1;

 fail:
  free(buf);
  return err;
}


int cli_enc(int argc, char **argv)
{
  uint8_t *buf;
  ssize_t slen;
  size_t clen;
  size_t len;
  size_t max;
  size_t h;
  size_t b;
  size_t d;
  int err;

  if ((4 > argc) || (MAX_LOCAL_ARGV < argc)) {
    fprintf(stderr, "Usage: %s <nick 1> .. <nick n> <file in> <file out>\n", argv[0]);
    return 0;
  }
  slen = file_size(argv[argc-2]);

  if (0 > slen) {
    fprintf(stderr, "Can't open file: %s\n", strerror(errno));
    return IO_ERROR;
  }
  /* Note: err below cannot become negative because max is small and positive.
   */
  len = (size_t)slen;
  max = argc - 3;
  err = create_efile_helper(len, max, &clen, &h, &b, &d);
  buf = (uint8_t *)malloc(clen);

  if (NULL == buf) {
    fprintf(stderr, "Can't allocate memory for file.\n");
    return -1;
  }
  err = file_loadsync(argv[argc-2], &buf[d], len);

  if (0 > err) {
    fprintf(stderr, "Error loading file: %s\n", strerror(errno));
    goto fail;
  }
  err = create_efile(buf, clen, &argv[1], max, h, b);

  if (0 > err) {
    goto fail;
  }
  err = file_savesync(argv[argc-1], buf, clen);

  if (0 > err) {
    fprintf(stderr, "Error saving file: %s\n", strerror(errno));
    goto fail;
  }
  err = 1;

 fail:
  free(buf);
  return err;
}


int cli_restore(int argc, char **argv)
{
  key_entry_t *e;
  uint8_t *buf;
  ssize_t slen;
  size_t offs;
  size_t len;
  int err;

  if (2 != argc) {
    fprintf(stderr, "Usage: %s <file in>\n", argv[0]);
    return 0;
  }
  slen = file_size(argv[1]);

  if (0 > slen) {
    fprintf(stderr, "Can't open file: %s\n", strerror(errno));
    return IO_ERROR;
  }
  len = (size_t)slen;
  buf = (uint8_t *)malloc(len);

  if (NULL == buf) {
    fprintf(stderr, "Can't allocate memory.\n");
    return MEMORY_ERROR;
  }
  /* From here on, set err, exit through goto fail1.
   */
  err = -1;

  if (0 > file_loadsync(argv[1], buf, len)) {
    fprintf(stderr, "Error loading file: %s\n", strerror(errno));
    goto fail1;
  }
  if (0 > open_efile(buf, len, &e, &offs)) {
    goto fail2;
  }
  if (0 > key_store_match((key_store_t *)&buf[offs], len - offs)) {
    goto fail2;
  }
  if (0 > key_store_merge((key_store_t *)&buf[offs])) {
    goto fail2;
  }
  err = 1;

 fail2:
  memset(buf, 0, len);
  
 fail1:
  free(buf);
  return err;
}


int cli_dec(int argc, char **argv)
{
  key_entry_t *e;
  uint8_t *buf;
  ssize_t slen;
  size_t offs;
  size_t len;
  int err;

  if (3 != argc) {
    fprintf(stderr, "Usage: %s <file in> <file out>\n", argv[0]);
    return 0;
  }
  slen = file_size(argv[1]);

  if (0 > slen) {
    fprintf(stderr, "Can't open file: %s\n", strerror(errno));
    return IO_ERROR;
  }
  len = (size_t)slen;
  buf = (uint8_t *)malloc(len);

  if (NULL == buf) {
    fprintf(stderr, "Can't allocate memory.\n");
    return MEMORY_ERROR;
  }
  /* From here on, set err, exit through goto fail1.
   */
  err = file_loadsync(argv[1], buf, len);

  if (0 > err) {
    fprintf(stderr, "Error loading file: %s\n", strerror(errno));
    goto fail1;
  }
  err = open_efile(buf, len, &e, &offs);

  if (0 > err) {
    goto fail2;
  }
  err = file_savesync(argv[2], &buf[offs], len - offs);

  if (0 > err) {
    fprintf(stderr, "Error saving file: %s\n", strerror(errno));
    goto fail2;
  }
  err = 1;

 fail2:
  memset(buf, 0, len);
  
 fail1:
  free(buf);
  return err;
}

