
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/param.h>
#include <sys/wait.h>
#include "file.h"
#include "ee.h"
#include "viz.h"
#include "encode.h"
#include "edit.h"
#include "keystore.h"
#include "cli.h"
#include "devurandom.h"
#include "tweetnacl.h"
#include "zeit.h"
#include "customization.h"
#include "cli_keys.h"

extern key_store_t *keys;


int cli_uncheck(int argc, char **argv)
{
  key_entry_t *e;
  int err;

  if (2 != argc) {
    fprintf(stderr, "Usage: %s <nick>\n", argv[0]);
    return 0;
  }
  err = find_nick(argv[1], &e);

  if (NICK_UNKNOWN == err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  if (0 == (e->flags & FLAG_EM)) {
    fprintf(stderr, "I don't have a key for nick %s.\n", argv[1]);
    return -1;
  }
  e->flags = e->flags & (~FLAG_CHECKED);
  return 1;
}


int cli_check(int argc, char **argv)
{
  key_entry_t *e;
  uint32_t m;
  uint32_t n;
  uint32_t r;
  uint32_t x;
  uint32_t i;
  uint8_t text[LEN_PK + LEN_PK];
  /* 
   * This would be required if we chosen slightly better probabilities and
   * the computational security model. See also below.
   *
   * uint8_t hash[64];
   */
  int err;
  int s;

  if ((2 > argc) || (3 < argc)) {
    fprintf(stderr, "Usage: %s <nick> [<integer> | ok]\n", argv[0]);
    return 0;
  }
  err = find_nick(argv[1], &e);

  if (NICK_UNKNOWN == err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  if (0 == (e->flags & FLAG_EM)) {
    fprintf(stderr, "I don't have a key for nick %s.\n", argv[1]);
    return -1;
  }
  if (3 == argc) {
    /*
     * See if we are asked to simply check this nick as verified.
     */
    if (strncmp(argv[2], "ok\0", 3) == 0) {
      e->flags = e->flags | FLAG_CHECKED;
      return 1;
    }
    /* No, we are asked to do a verification.
     */
    i = (uint32_t)atoi(argv[2]);

    /* This is enough because i is unsigned.
     */
    if (0xffff < i) {
      fprintf(stderr, "I expected an integer in the range [0:65536].\n");
      return -1;
    }
    n = (uint32_t)i;
  }
  else {
    /* No, this does not run where sizeof(int) == 1.
     */
    randombytes((uint8_t *)&m, 4);
    
    /* We use a range that is a power of two; using a range that is a power
     * of 10 would lead to a bias.
     */
    n = m % CHECK_PRIME;
  }
  /* We use a canonical ordering of public keys as input into the key hash.
   */
  s = memcmp(e->me.pk, e->em.pk, LEN_PK);

  if (0 > s) {
    memmove(text, e->em.pk, LEN_PK);
    memmove(&text[LEN_PK], e->me.pk, LEN_PK);
  }
  else {
    memmove(text, e->me.pk, LEN_PK);
    memmove(&text[LEN_PK], e->em.pk, LEN_PK);
  }
  /* One approach is to use only the first 32 octets of the hash of both
   * public keys.  This halves the degree of the Reed Solomon polynomial
   * while keeping the space of coefficients the same.  This translates to
   * better security on the one hand.  On the other hand, we loose some
   * security due to the collision probability of the hash function.  Worse,
   * we loose the guarantee of security against unbounded adversaries.
   * After some deliberation, I go for the unbounded adversary model.
   */
  // crypto_hash(hash, text, sizeof(text));

  /* Our modulus for checking is slightly larger than 2^16, we operate on
   * two octets at a time.  If we chose the computational security model and
   * encodings based on a hash of two keys then the loop would have to use
   * 32 and the array would have to be the hash.
   */
  for (r=0, x=1, i=0; i<sizeof(text); i += 2) {
    m = ((uint32_t)text[i]) << 8;
    m = m | ((uint32_t)text[i+1]);
    r = (r + m * x) % CHECK_PRIME;
    x = (n * x)     % CHECK_PRIME;
  }
  printf("    For nick: %s\n", e->nick);
  printf("   With tags: %s\n\n", e->tags);

  if (3 != argc) {
    printf("  Check with: %d\n", n);
  }
  printf(EES_CMP, r);
  printf("You are in trouble if the comparison fails!\n");
  return 1;
}


int cli_transfer(int argc, char **argv)
{
  key_entry_t *a;
  key_entry_t *b;
  update_t upd;
  nonce_t nonce;
  uint8_t buf[base64_length(sizeof(update_t))];
  size_t len;
  int err;
  int fd;

  if ((3 > argc) || (4 < argc)) {
    fprintf(stderr, "Usage: %s <from nick> <to nick> [<file>]\n", argv[0]);
    return 0;
  }
  err = find_nick(argv[1], &a);

  if (0 > err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  err = find_nick(argv[2], &b);

  if (0 > err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[2]);
    return -1;
  }
  if (a == b) {
    fprintf(stderr, "A nick can't transfer to itself.\n");
    return -1;
  }
  err = create_update(&upd, &nonce, a, b);

  if (0 > err) {
    fprintf(stderr, "Ask nick %s for a key, first.\n", argv[1]);
    return -1;
  }
  if (3 == argc) {
    len = encode64(buf, (uint8_t *)&upd, sizeof(update_t));
    update_print(STDOUT_FILENO, buf, len, 0, 0);
    return 1;
  }
  fd = open(
    argv[3],
    O_WRONLY | O_CREAT | O_NOFOLLOW,
    S_IRUSR  | S_IWUSR | S_IRGRP
  );

  if (0 > fd) {
    fprintf(stderr, "Can't open file: %s\n", strerror(errno));
    return -1;
  }
  len = encode64(buf, (uint8_t *)&upd, sizeof(update_t));
  update_print(fd, buf, len, 0, 0);

  if (close(fd) < 0) {
    fprintf(stderr, "Error closing file: %s\n", strerror(errno));
    return -1;
  }
  return 1;
}


int cli_confirm(int argc, char **argv)
{
  key_entry_t *a;
  key_entry_t *b;
  nonce_t nonce;
  uint8_t buf[base64_length(sizeof(update_t))];
  size_t len;
  int err;
  int fd;

  if ((3 > argc) || (4 < argc)) {
    fprintf(stderr, "Usage: %s <from nick> <to nick> [<file>]\n", argv[0]);
    return 0;
  }
  err = find_nick(argv[1], &a);

  if (0 > err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  err = find_nick(argv[2], &b);

  if (NICK_UNKNOWN == err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[2]);
    return -1;
  }
  if (a == b) {
    fprintf(stderr, "A nick can't confirm to itself.\n");
    return -1;
  }
  if (3 == argc) {
    len = distill64(STDIN_FILENO, buf, sizeof(buf));
    len = decode64(buf, len);
  }
  else {
    fd = open(argv[3], O_RDONLY);

    if (0 > fd) {
      fprintf(stderr, "Can't open file: %s\n", strerror(errno));
      return -1;
    }
    len = distill64(fd, buf, sizeof(buf));
    len = decode64(buf, len);

    if (close(fd) < 0) {
      fprintf(stderr, "Error closing file: %s\n", strerror(errno));
      return -1;
    }
  }
  if (sizeof(update_t) != len) {
    fprintf(stderr, "Wrong input length, did you paste correctly?\n");
    return -1;
  }
  err = open_update((update_t *)buf, &nonce, a, b);

  if (KEY_UNKNOWN == err) {
    fprintf(stderr, "Ask %s to send a key, first.\n", argv[1]);
    return -1;
  }
  if (0 > err) {
    fprintf(stderr, EES_CONFNO, argv[1], argv[2]);
    return -1;
  }
  fprintf(stdout, EES_CONFYES, argv[1], argv[2]);
  return 1;
}


int cli_export(int argc, char **argv)
{
  key_entry_t *e;
  uint8_t buf[base64_length(LEN_PK)];
  size_t len;
  int err;
  int fd;

  if ((2 > argc) || (3 < argc)) {
    fprintf(stderr, "Usage: %s <nick> [<file>]\n", argv[0]);
    return 0;
  }
  err = find_nick(argv[1], &e);

  if (NICK_UNKNOWN == err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  if (2 == argc) {
    len = encode64(buf, e->me.pk, LEN_PK);
    key_print(STDOUT_FILENO, buf, len, 0, 1);
    return 1;
  }
  len = encode64(buf, e->me.pk, LEN_PK);
  fd  = open(
    argv[2],
    O_WRONLY | O_CREAT | O_NOFOLLOW,
    S_IRUSR  | S_IWUSR | S_IRGRP
  );

  if (0 > fd) {
    fprintf(stderr, "Can't open file: %s\n", strerror(errno));
    return -1;
  }
  key_print(fd, buf, len, 0, 1);
  
  if (close(fd) < 0) {
    fprintf(stderr, "Error closing file: %s\n", strerror(errno));
    return -1;
  }
  return 1;
}


int cli_import(int argc, char **argv)
{
  key_entry_t *e;
  uint8_t buf[base64_length(LEN_PK)];
  size_t len;
  int err;
  int fd;

  if ((2 > argc) || (3 < argc)) {
    fprintf(stderr, "Usage: %s <nick> [<file>]\n", argv[0]);
    return 0;
  }
  err = find_nick(argv[1], &e);

  if (0 > err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  if (2 == argc) {
    len = distill64(STDIN_FILENO, buf, sizeof(buf));
    len = decode64(buf, len);
  }
  else {
    fd = open(argv[2], O_RDONLY);

    if (0 > fd) {
      fprintf(stderr, "Can't open file: %s\n", strerror(errno));
      return -1;
    }
    len = distill64(fd, buf, sizeof(buf));
    len = decode64(buf, len);

    if (close(fd) < 0) {
      fprintf(stderr, "Error closing file: %s\n", strerror(errno));
      return -1;
    }
  }
  if (LEN_PK != len) {
    fprintf(stderr, "Wrong input length, did you paste correctly?\n");
    return -1;
  }
  err = set_em(&e, buf);

  if (KEY_EXISTS == err) {
    if (memcmp(e->em.pk, buf, LEN_PK) == 0) {
      fprintf(stderr, "I already imported this key for nick %s.\n", argv[1]);
      return 1;
    }
    fprintf(stderr, "I already have a different key for nick %s.\n", argv[1]);
    return -1;
  }
  if (KEY_ME == err) {
    fprintf(stderr, "Careful, this is your own key for nick %s.\n", e->nick);
    return -1;
  }
  if (KEY_IN_USE == err) {
    fprintf(stderr, "Careful, that key is from nick %s.\n", e->nick);
    return -1;
  }
  return 1;
}


