
#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <ctype.h>
#include <sys/mman.h>
#include <sys/param.h>
#include <sys/wait.h>
#include "file.h"
#include "ee.h"
#include "viz.h"
#include "encode.h"
#include "edit.h"
#include "keystore.h"
#include "cli.h"
#include "devurandom.h"
#include "zeit.h"
#include "customization.h"
#include "cli_nicks.h"

extern key_store_t *keys;

int cli_info(int argc, char **argv)
{
  key_entry_t *e;
  uint8_t buf[base64_length(LEN_PK)+1];
  int err;
  int i;

  if (2 != argc) {
    fprintf(stderr, "Usage: %s <nick>\n", argv[0]);
    return 0;
  }
  err = find_nick(argv[1], &e);

  if (0 > err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  fprintf(stdout, "%10s : %s\n", "nick", e->nick);
  fprintf(stdout, "%10s : %s\n", "tags", e->tags);
  fprintf(stdout, "%10s : %s\n", "added", zeit_ctime(e->stat_date));
  fprintf(stdout, "%10s : %d\n", "sent", e->stat_enc);
  fprintf(stdout, "%10s : %d\n", "received", e->stat_dec);

  i = encode64(buf, e->me.pk, LEN_PK);
  buf[i] = '\0';
  fprintf(stdout, "%10s : %s\n", "my pk", buf);

  i = encode64(buf, e->em.pk, LEN_PK);
  buf[i] = '\0';
  fprintf(stdout, "%10s : %s\n", "his/her pk", buf);
  return 1;
}


uint32_t cli_show_keys(uint16_t mask, uint16_t flags, const char *mark, char *prefix)
{
  key_entry_t *e;
  uint32_t count;
  uint32_t i;
  size_t n;

  n = strnlen(prefix, NICK_LENGTH);

  for (count=0, i=0; i < keys->count; i++) {
    e = &keys->entry[i];

    if (strncasecmp(e->nick, prefix, n) != 0) {
      continue;
    }
    if ((e->flags & mask) == flags) {
      printf(EES_SHOW, e->nick, mark, e->tags);
      count++;
    }
  }
  return count;
}


int cli_show(int argc, char **argv)
{
  uint32_t checked;
  uint32_t ready;
  uint32_t other;
  uint16_t mask;
  uint32_t sum;
  char *prefix;

  if (2 < argc) {
    fprintf(stderr, "Usage: %s [<prefix>]\n", argv[0]);
    return 0;
  }
  if (0 == keys->count) {
    printf("No nicks added so far.\n");
    return 1;
  }
  checked = 0;
  ready   = 0;
  mask    = FLAG_EM | FLAG_CHECKED;

  if (2 == argc) {
    prefix = argv[1];
  }
  else {
    prefix = "";
  }
  // printf("\n");

  checked = cli_show_keys(mask, FLAG_EM | FLAG_CHECKED, EES_CK,  prefix);
  ready   = cli_show_keys(mask, FLAG_EM,                EES_K,   prefix);
  other   = cli_show_keys(mask, 0,                      EES_NEW, prefix);

  // printf("\n");

  sum = checked + ready;

  if (0 < sum) {
    printf(
      "%d of %d (%u%%) nicks verified, %d nicks open.\n",
      checked,
      sum,
      (checked * 100) / sum,
      other
    );
  }
  return 1;
}


int cli_add(int argc, char **argv)
{
  key_entry_t *e;
  uint8_t hash[LEN_HASH];
  char pw[PW_LENGTH];
  size_t n;
  size_t i;
  char *s;
  int err;

  if ((2 > argc) || (3 < argc)) {
    fprintf(stderr, "Usage: %s <nick> <tag string>\n", argv[0]);
    fprintf(stderr, "       %s private <tag string>\n", argv[0]);
    return 0;
  }
  s = "";
  n = strlen(argv[1]);

  if (NICK_LENGTH < n) {
    fprintf(stderr, "Nick %lu characters too long.\n", n-NICK_LENGTH);
    return -1;
  }
  if (3 == argc) {
    n = strlen(argv[2]);

    if (TAGS_LENGTH < n) {
      fprintf(stderr, "Tag string %lu characters too long.\n", n-TAGS_LENGTH);
      return -1;
    }
    s = argv[2];
  }
  err = create_nick(argv[1], s, &e);

  if (NICK_EXISTS == err) {
    fprintf(stderr, "Nick %s already exists.\n", argv[1]);
    return -1;
  }
  if (STORE_OVERFLOW == err) {
    fprintf(stderr, "Keys file is maxed out. Ask a wizard.\n");
    return -1;
  }
  /* If the nick name is "private" then we create an entry that is fully
   * initialized and can be used to encrypt without importing another public
   * key.  These keys can be used to encrypt data to oneself.
   */
  if (strncmp(argv[1], "private\0", 8) == 0) {
    /*
     * Copy the current me.pk to em.pk, create new me.pk and me.sk.  The
     * previous me.sk is lost und unrecoverable.
     */
    memcpy(e->em.pk, e->me.pk, sizeof(e->em.pk));
    crypto_box_keypair(e->me.pk, e->me.sk);
    e->flags = e->flags | FLAG_EM | FLAG_CHECKED;
    return 1;
  }
  /* If the nick name is "passwd" then we deterministically create an entry
   * that is fully initialized based on a user-provided password.  This
   * entry can be used to lock and unlock one's key store manually by means
   * of a backup, destroy, restore cycle.  It can be used to perform local
   * backups as well.  We leverage the fact that all strings are mapped to
   * scalars and group elements in tweetnacl.
   */
  if (strncmp(argv[1], "passwd\0", 6) == 0) {
    printf("Please enter a password (max. %d chars): ", PW_LENGTH);
    i = edit_passwd(pw, PW_LENGTH);

    crypto_hash(hash, (uint8_t *)pw, i);
    memcpy(e->me.sk, hash, LEN_SK);
    crypto_scalarmult_base(e->em.pk, e->me.sk);

    crypto_hash(hash, hash, sizeof(hash));
    memcpy(e->me.sk, hash, LEN_SK);
    crypto_scalarmult_base(e->me.pk, e->me.sk);

    e->flags = e->flags | FLAG_EM | FLAG_CHECKED;
  }
  return 1;
}


int cli_rename(int argc, char **argv)
{
  key_entry_t *e1;
  key_entry_t *e2;
  int err;

  if ((3 > argc) || (4 < argc)) {
    fprintf(stderr, "Usage: %s <old nick> <new nick> [<new tag string>]\n", argv[0]);
    return 0;
  }
  err = find_nick(argv[1], &e1);

  if (0 > err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  err = find_nick(argv[2], &e2);

  if (0 <= err) {
    if (e1 != e2) {
      fprintf(stderr, "Nick %s already exists.\n", argv[2]);
      return -1;
    }
  }
  if (4 == argc) {
    memset( e1->tags,       0, sizeof(e1->tags));
    strncpy(e1->tags, argv[3], sizeof(e1->tags)-1);
  }
  memset( e1->nick,       0, sizeof(e1->nick));
  strncpy(e1->nick, argv[2], sizeof(e1->nick)-1);
  return 1;
}


int cli_clone(int argc, char **argv)
{
  key_entry_t *e1;
  key_entry_t *e2;
  char *s;
  int err;

  if ((3 > argc) || (4 < argc)) {
    fprintf(stderr, "Usage: %s <nick> <new nick> [<new tag string>]\n", argv[0]);
    return 0;
  }
  s   = (4 == argc) ? argv[3] : "";
  err = find_nick(argv[1], &e1);

  if (0 > err) {
    fprintf(stderr, "Nick %s is unknown.\n", argv[1]);
    return -1;
  }
  err = create_nick(argv[2], s, &e2);

  if (NICK_EXISTS == err) {
    fprintf(stderr, "Nick %s already exists.\n", argv[2]);
    return -1;
  }
  if (STORE_OVERFLOW == err) {
    fprintf(stderr, "Keys file is maxed out. Ask a wizard.\n");
    return -1;
  }
  memcpy(&e2->me, &e1->me, sizeof(me_t));
  return 1;
}


int cli_delete(int argc, char **argv)
{
  int i;

  if ((2 > argc) || (MAX_LOCAL_ARGV < argc)) {
    fprintf(stderr, "Usage: %s <nick 1> ... <nick n>\n", argv[0]);
    return 0;
  }
  for (i=1; i<argc; i++) {
    if (0 > delete_nick(argv[i])) {
      fprintf(stderr, "Nick %s is unknown.\n", argv[i]);
    }
  }
  return 1;
}

