
#include <sys/socket.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <ctype.h>
#include "edit.h"
#include "list.h"

/* The general principle of the implementation is based on an idea
 * implemented in emacs, based on what I read somewhere.  Instead of editing
 * a line, one edits a "hole" in a buffer.  The beginning of the "hole" is
 * where the current cursor position is.  Characters left of the cursor are
 * flush against the beginning of the buffer.  Characters right of the
 * cursor are flush against the buffer's end.
 *
 * For comparison, when editing a line using the naive approach, whenever a
 * character is inserted, one needs to shift the characters following the
 * insertion point to the right.  In our approach, inserting characters
 * simply writes characters into the hole which gets smaller with each
 * inserted character.
 *
 * When moving the cursor, single characters get copied from either end of
 * the hole to the other end.  The most expensive operation is "jumping" the
 * cursor position to arbitrary positions in the buffer because a string of
 * characters needs to be copied from one end to the other end of the hole.
 */

void edit_print_tail(edit_t *eb, int del)
{
  char buf[LOG_LINE+4];
  int pos;
  int len;
  int n;

  pos = eb->pos + eb->len;
  len = MAX_LINE - pos;
  
  if (0 < len) {
    write(STDOUT_FILENO, &eb->buf[pos], len);
  }
  if (1 == del) {
    write(STDOUT_FILENO, " ", 1);
    len++;
  }
  if (0 < len) {
    n = snprintf(buf, sizeof(buf), "\033[%dD", len);
    write(STDOUT_FILENO, buf, n);
  }
}


void edit_kill_to_end(edit_t *eb)
{
  int pos;
  int len;

  pos = eb->pos + eb->len;
  len = MAX_LINE - pos;
  
  if (0 == len) {
    return;
  }
  eb->len += len;
  write(STDOUT_FILENO, "\033[0K", 4);
}


void edit_delete(edit_t *eb)
{
  int pos;
  int len;

  pos = eb->pos + eb->len;
  len = MAX_LINE - pos;

  if (0 == len) {
    return;
  }
  eb->len++;
  edit_print_tail(eb, 1);
}


void edit_backspace(edit_t *eb)
{
  if (0 == eb->pos) {
    return;
  }
  eb->pos--;
  eb->len++;

  write(STDOUT_FILENO, "\033[D \033[D", 7);
  edit_print_tail(eb, 1);
}


void edit_insert(edit_t *eb, char c)
{
  if (0 == eb->len) {
    return;
  }
  eb->buf[eb->pos] = c;
  eb->pos++;
  eb->len--;

  write(STDOUT_FILENO, &c, 1);
  edit_print_tail(eb, 0);
}


void edit_move_backward(edit_t *eb)
{
  if (0 == eb->pos) {
    return;
  }
  eb->pos--;
  eb->buf[eb->pos+eb->len] = eb->buf[eb->pos];

  write(STDOUT_FILENO, "\033[D", 3);
}


void edit_move_forward(edit_t *eb)
{
  if (MAX_LINE == (eb->pos + eb->len)) {
    return;
  }
  eb->buf[eb->pos] = eb->buf[eb->pos+eb->len];
  eb->pos++;

  write(STDOUT_FILENO, &eb->buf[eb->pos-1], 1);
}


/* Beware the invariant that n >= 0
 */
void edit_move_to(edit_t *eb, int n)
{
  char buf[LOG_LINE+4];
  int diff;

  if (n > MAX_LINE - eb->len) {
    return;
  }
  /* Move to the left */
  if (n < eb->pos) {
    diff = eb->pos - n;
    memmove(&eb->buf[eb->pos+eb->len-diff], &eb->buf[n], diff);
    eb->pos = n;

    n = snprintf(buf, sizeof(buf), "\033[%dD", diff);
    write(STDOUT_FILENO, buf, n);
    return;
  }
  /* Move to the right */
  if (n > eb->pos) {
    diff = n - eb->pos;
    memmove(&eb->buf[eb->pos], &eb->buf[eb->pos+eb->len], diff);
    eb->pos = n;

    n = snprintf(buf, sizeof(buf), "\033[%dC", diff);
    write(STDOUT_FILENO, buf, n);
  }
}


void edit_hist_cpy(edit_t *eb)
{
  edit_he_t *he;

  edit_move_to(eb, 0);
  edit_kill_to_end(eb);

  he = (edit_he_t *)eb->here;

  if (NULL != he) {
    strncpy(eb->buf, he->buf, he->len);
    eb->pos = he->len;
    eb->len = MAX_LINE - eb->pos;

    write(STDOUT_FILENO, eb->buf, eb->pos);
  }
}


void edit_hist_up(edit_t *eb)
{
  lnod_t *ln;

  ln = eb->here;

  /* Are we at the bottom?
   */
  if (NULL == ln) {
    ln = eb->hist.last;
  }
  else {
    ln = ln->pred;
  }
  /* Are we already at the top?
   */
  if (NULL == ln->pred) {
    write(STDOUT_FILENO, "\007", 1);
    return;
  }
  eb->here = ln;
  edit_hist_cpy(eb);
}


void edit_hist_down(edit_t *eb)
{
  lnod_t *ln;

  ln = eb->here;

  /* Are we already at the bottom?
   */
  if (NULL == ln) {
    write(STDOUT_FILENO, "\007", 1);
    return;
  }
  /* Move one down and check whether there is a valid item.
   */
  ln = ln->succ;

  if (NULL == ln->succ) {
    ln = NULL;
  }
  eb->here = ln;
  edit_hist_cpy(eb);
}


void edit_newline(edit_t *eb)
{
  edit_he_t *he;
  char *buf;
  int n;

  /* Move to the end of the line and then print a newline.
   */
  n = MAX_LINE - eb->len;
  edit_move_to(eb, n);
  eb->buf[n] = 0;

  write(STDOUT_FILENO, "\n", 1);

  /* Empty lines need not go into the history.
   */
  if (0 >= n) {
    return;
  }
  /* Allocate a history entry and save what has been entered.
   */
  he  = (edit_he_t *)calloc(1, sizeof(edit_he_t));
  buf = (char *)malloc(n);

  if ((NULL == he) || (NULL == buf)) {
    fprintf(stderr, "Can't allocate memory for command-line history.");
    exit(-1);
  }
  strncpy(buf, eb->buf, n);

  /* Setting eb->here = NULL moves us to the end of the history.
   */
  he->buf  = buf;
  he->len  = n;
  eb->here = NULL;

  list_add_tail(&eb->hist, &he->ln);
}


void edit_reset(edit_t *eb)
{
  struct winsize ws;

  ioctl(STDOUT_FILENO, TIOCGWINSZ, &ws);
  eb->row = ws.ws_row;
  eb->col = ws.ws_col;

  eb->buf[0] = 0;
  eb->pos    = 0;
  eb->len    = MAX_LINE;
}


void edit_init(edit_t *eb)
{
  edit_reset(eb);
  list_init(&eb->hist);
  eb->here = NULL;
}


void term_raw(tcflag_t *save)
{
  struct termios t;

  tcgetattr(STDIN_FILENO, &t);
  *save = t.c_lflag;
  t.c_lflag &= ~((tcflag_t)(ICANON | ECHO));
  tcsetattr(STDIN_FILENO, TCSANOW, &t);
}


void term_restore(tcflag_t *save)
{
  struct termios t;

  tcgetattr(STDIN_FILENO, &t);
  t.c_lflag = *save;
  tcsetattr(STDIN_FILENO, TCSANOW, &t);
}


int edit_vt100_esc(edit_t *eb)
{
  char c;
  int val[2];
  int i;
  int n;

  if (read(STDIN_FILENO, &c, 1) != 1) {
    return -1;
  }
  if ('[' != c) {
      return -1;
  }
  n = 0;
  i = 0;

  while (read(STDIN_FILENO, &c, 1) == 1) {
    if (isdigit(c)) {
      n = n * 10 + (c - '0');
      continue;
    }
    if (';' == c) {
      val[i] = n;
      i = 1 - i;
      n = 0;
      continue;
    }
    if ('R' == c) {
      eb->y = n;
      eb->x = val[1-i];
      return -1;
    }
    break;
  }
  return c;
}


void edit_line(edit_t *eb, char *prompt)
{
  char c;
  int n;

  term_raw(&eb->save);

  /* Line wrap on */
  write(STDOUT_FILENO, "\033[?7h", 5);

  /* Prompt */
  write(STDOUT_FILENO, prompt, strlen(prompt));

  /* Request position report */
  write(STDOUT_FILENO, "\033[6n", 4);

  while (read(STDIN_FILENO, &c, 1) == 1) {
    if ('\n' == c) {
      break;
    }
    if ('\177' == c) {
      edit_backspace(eb);
      continue;
    }
    if ('\004' == c) {
      edit_delete(eb);
      continue;
    }
    if ('\001' == c) {
      edit_move_to(eb, 0);
      continue;
    }
    if ('\005' == c) {
      edit_move_to(eb, MAX_LINE - eb->len);
      continue;
    }
    if ('\013' == c) {
      edit_kill_to_end(eb);
      continue;
    }
    if (isprint(c)) {
      edit_insert(eb, c);
      continue;
    }
    if (0x1b != c) {
      //printf("0x%x\n", c);
      continue;
    }
    n = edit_vt100_esc(eb);

    if (0 > n) {
      continue;
    }
    c = (char)n;

    if ('D' == c) {
      edit_move_backward(eb);
      continue;
    }
    if ('C' == c) {
      edit_move_forward(eb);
      continue;
    }
    if ('A' == c) {
      edit_hist_up(eb);
      continue;
    }
    if ('B' == c) {
      edit_hist_down(eb);
      continue;
    }
  }
  edit_newline(eb);
  term_restore(&eb->save);
}


int edit_passwd(char *pw, size_t len)
{
  char animation[] = "-\\|/-\\|/";
  tcflag_t flag;
  size_t i;
  char c;
  int n;

  term_raw(&flag);
  memset(pw, 0, len);

  i = 0;
  n = getchar();

  while ((EOF != n) && (0x0a != n)) {
    c = (char)n;

    if ((0 < i) && ('\177' == c)) {
      i--;
      pw[i] = '\0';
      write(STDOUT_FILENO, &animation[i % 8], 1);
      write(STDOUT_FILENO, "\033[D", 3);
    }
    else if ((i < len) && (isprint(c))) {
      pw[i] = c;
      i++;
      write(STDOUT_FILENO, &animation[i % 8], 1);
      write(STDOUT_FILENO, "\033[D", 3);
    }
    else {
      write(STDOUT_FILENO, "\007", 1);
    }
    n = getchar();
  }
  term_restore(&flag);
  write(STDOUT_FILENO, " \n", 2);
  return i;
}


int edit_parse(char *buf, int *argc, char **argv, int max)
{
  char *p;
  char c;
  int n;
  int i;

  p = buf;
  n = 0;
  i = 0;
  c = *buf++;

  while (0 != c) {
    switch(n) {
    case 0:
      /* Reading spaces
       */
      if (isspace(c)) {
	break;
      }
      if (i >= max) {
	return -2;
      }
      argv[i++] = p;

      /* Deciding on first non-space char
       */
      if ('\"' == c) {
	n = 2;
      }
      else if ('\\' == c) {
	n = 3;
      }
      else {
	*p++ = c;
	n = 4;
      }
      break;

    case 2:
      /* Reading quoted string
       */
      if ('\"' == c) {
	*p++ = 0;
	n = 5;
	break;
      }
      *p++ = c;
      break;

    case 3:
      /* Reading escaped char
       */
      *p++ = c;
      n = 4;
      break;

    case 4:
      /* Reading string
       */
      if ('\\' == c) {
	n = 3;
	break;
      }
      if ('\"' == c) {
	return -1;
      }
      if (isspace(c)) {
	*p++ = 0;
	n = 0;
	break;
      }
      *p++ = c;
      break;

    case 5:
      /* Forced space
       */
      if (isspace(c)) {
	n = 0;
	break;
      }
      return -1;

    default:
      return -1;
    }
    c = *buf++;
  }
  if ((2 == n) || (3 == n)) {
    return -1;
  }
  *p = 0;
  *argc = i;
  return 1;
}
