
#include <errno.h>
#include <sys/socket.h>
#include <string.h>
#include <unistd.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <fcntl.h>
#include <pwd.h>
#include "file.h"
#include "zeit.h"
#include "ee.h"
#include "cli.h"
#include "keystore.h"
#include "customization.h"

extern key_store_t *keys;

char disclaimer[] =
"\n\
\tCopyright (c) 2015, Volker Roth <volker.roth@fu‐berlin.de>\n\
\tAll rights reserved. Please see file COPYING for licensing\n\
\tinformation and disclaimers.\n\n\
";


void usage(char *name)
{
  printf("USAGE: %s <flags>\n", name);
  printf("  -h        this message\n");
  printf("  -s <int>  sets the decoration theme\n");
  printf("  -a <name> loads alternative keystore\n");

#ifdef EE_UPGRADABLE
  printf("  -u <name> upgrade and save keystore\n");
#endif
}


void checkname(char *name)
{
  int i;

  if (NULL == name) {
    return;
  }
  for (i=strlen(name)-1; i >= 0; i--) {
    if (!isalnum(name[i])) {
      fprintf(stderr, "Illegal store name, only alphanumerics are allowed.");
      exit(-1);
    }
  }
}


int main(int argc, char **argv)
{
  /* For argument parsing.
   */
  char *storename;
  char *name;
  int ch;

#ifdef EE_UPGRADABLE
  char *upgrade;
  upgrade = NULL;
#endif

  storename = "1";
  name = argv[0];

  set_style(0);

  while ((ch = getopt(argc, argv, "ha:s:")) != -1) {
    switch (ch)
    {
    case 's':
      set_style(atoi(optarg));
      break;

    case 'a':
      storename = strdup(optarg);
      break;

#ifdef EE_UPGRADABLE
    case 'u':
      upgrade = strdup(optarg);
      break;
#endif
      
    case 'h':
    case '?':
      usage(name);
      exit(0);

    default:
      break;
    }
  }
  argc -= optind;
  argv += optind;

  if (argc != 0) {
    usage(name);
    exit(0);
  }
  checkname(storename);

#ifdef EE_UPGRADABLE
  checkname(upgrade);
#endif

  struct passwd *pwd;
  uint64_t now;
  int cwd;

  printf("%s", disclaimer);

  /* Printing out system time
   */
  now = zeit_now();

  printf(
    "\tThis is ee version: %u.%u\n\tKeystore version  : %u\n",
    EE_VERSION_MAJOR,
    EE_VERSION_MINOR,
    EE_VERSION_KEYSTORE
  );
  printf(EES_TIME, zeit_ctime(now));

  if (geteuid() != 0) {
    fprintf(stderr, "%s", EES_PRIV);
  }
  pwd = getpwuid(getuid());
  
  if (NULL == pwd) {
    fprintf(stderr, "Error, I cannot get info for uid %d.\n", getuid());
    exit(-1);
  }
  cwd = open(".", O_RDONLY);

  if (0 > cwd) {
    fprintf(stderr, "Error, I cannot get the CWD.\n");
    exit(-1);
  }
  if (chdir(pwd->pw_dir) == -1) {
    fprintf(stderr, "Error, I cannot change into %s.\n", pwd->pw_dir);
    exit(-1);
  }
  if (chdir(".ee/secret/keys") == -1) {
    fprintf(stderr, "Error, I cannot change into .ee/secret/keys/\n");
    exit(-1);
  }
  /* Dropping privileges.  Make sure the order is correct, i.e., group
   * privileges are dropped first.
   */
  setgid(pwd->pw_gid);
  setuid(pwd->pw_uid);

  if (seteuid(0) >= 0) {
    fprintf(stderr, "Error, I couldn't drop privileges.\n");
    exit(-1);
  }
  if (0 > key_store_init(storename)) {
    fprintf(stderr, "Error, I cannot open or create the keystore.\n");
    exit(-1);
  }
  if (cli_checkpw(1) < 0) {
    exit(-1);
  }

#ifdef EE_UPGRADABLE
  /* Check whether we are asked to upgrade the keystore.
   */
  if (NULL != upgrade) {
    upgrade_keystore(upgrade);
  }
#endif

  fchdir(cwd);
  close(cwd);

  cli_main();
}
