
#ifndef EASY_ENCRYPT_H
#define EASY_ENCRYPT_H

#include "tweetnacl.h"

/* Used for checking keys.
 */
#define CHECK_PRIME 65537

/* Keystore error codes.  All error codes must be negative.
 */
#define NICK_UNKNOWN    -1
#define NICK_EXISTS     -2
#define KEY_EXISTS      -3
#define KEY_ME          -4
#define STORE_OVERFLOW  -5
#define KEY_UNKNOWN     -6
#define KEY_IN_USE      -7
#define CRYPTO_ERROR    -8
#define NOT_IMPLEMENTED -9
#define IO_ERROR        -10
#define MEMORY_ERROR    -11
#define BAD_FORMAT      -12
#define OTHER_ERROR     -13

/* Flags used in key entries and in keystore functions.
 */
#define FLAG_ME      1
#define FLAG_EM      2
#define FLAG_CHECKED 4

#define flag_has_me(e)     (((e)->flags & FLAG_ME)      == FLAG_ME)
#define flag_has_em(e)     (((e)->flags & FLAG_EM)      == FLAG_EM)
#define flag_is_checked(e) (((e)->flags & FLAG_CHECKED) == FLAG_CHECKED)

/* Lengths are chosen so that padding is avoided.
 */
#define KEYS_PER_STORE 128
#define NICK_LENGTH 20
#define TAGS_LENGTH 50
#define SEED_LENGTH 20
#define LIST_MAX 32
#define PW_LENGTH 64

/* Lengths derived from tweetnacl.
 */
#define LEN_PK    (crypto_box_PUBLICKEYBYTES)
#define LEN_SK    (crypto_box_SECRETKEYBYTES)
#define LEN_NONCE (crypto_box_NONCEBYTES)
#define LEN_MZERO (crypto_box_ZEROBYTES)
#define LEN_BZERO (crypto_box_BOXZEROBYTES)
#define LEN_KEY   (crypto_box_BEFORENMBYTES)
#define LEN_HASH  (crypto_hash_BYTES)

/* Lengths of basic building blocks in our structures.
 */
#define LEN_DATE 8
#define LEN_SEED 8
#define LEN_NPAD (LEN_NONCE - LEN_DATE - LEN_SEED)
#define LEN_KID  (LEN_PK >> 1)

typedef struct key_entry_t key_entry_t;
typedef struct key_store_t key_store_t;
typedef struct me_t me_t;
typedef struct em_t em_t;

struct me_t {
  uint8_t pk[LEN_PK];
  uint8_t sk[LEN_SK];
};

struct em_t {
  uint8_t pk[LEN_PK];
};

#define EE_VERSION_MAJOR 0
#define EE_VERSION_MINOR 3
#define EE_VERSION_KEYSTORE 1

/* Order of fields is chosen so that padding is avoided on ia64 platforms.
 */
struct key_entry_t {
  uint64_t stat_date;      /* Date when key was last changed */
  uint32_t stat_enc;
  uint32_t stat_dec;
  uint16_t flags;
  char     nick[NICK_LENGTH];
  char     tags[TAGS_LENGTH];
  me_t     me;
  em_t     em;
};

struct key_store_t {
  uint64_t arch;
  uint32_t version;
  uint32_t count;
  uint8_t  seed[SEED_LENGTH];
  uint8_t  pw[2];
  uint8_t  errors;
  uint8_t  flags;
  key_entry_t entry[KEYS_PER_STORE];
};

/* The envelope is designed to minimize space requirements while still
 * allowing to use the NaCl C API in a straightforward fashion, that is,
 * without moving arrays around.
 */

typedef struct envelope_txt_t envelope_txt_t;
typedef struct envelope_box_t envelope_box_t;
typedef struct envelope_t envelope_t;

struct envelope_txt_t {
  uint8_t date[LEN_DATE];       /* must be zeroed before unbox */
  uint8_t seed[LEN_SEED];       /* must be zeroed before unbox */
  uint8_t zero[LEN_MZERO];      /* must be zeroed before box */
  uint8_t key[LEN_KEY];         /* encrypted payload, the data encryption key */
  uint8_t id_send[LEN_KID];     /* sender's public key prefix */
  uint8_t id_recv[LEN_KID];     /* receiver's public key prefix */
};

struct envelope_box_t {
  uint8_t ctxt[sizeof(envelope_txt_t) - 2 * LEN_KID];
  uint8_t id_send[LEN_KID];
  uint8_t id_recv[LEN_KID];
};

struct envelope_t {
  union {
    envelope_txt_t txt;
    envelope_box_t box;
  } u;
};

typedef struct update_txt_t update_txt_t;
typedef struct update_bxi_t update_bxi_t;
typedef struct update_bxo_t update_bxo_t;
typedef struct update_t update_t;

struct update_txt_t {
  uint8_t date[LEN_DATE];       /* must be zeroed before unbox */
  uint8_t seed[LEN_SEED];       /* must be zeroed before unbox */
  uint8_t zero[LEN_MZERO];      /* must be zeroed before box */
  uint8_t pred[LEN_PK];         /* old pk */
  uint8_t succ[LEN_PK];         /* new pk */
};

struct update_bxo_t {
  uint8_t ctxt[sizeof(update_txt_t)];
};

struct update_bxi_t {
  uint8_t date[LEN_DATE];
  uint8_t seed[LEN_SEED];
  uint8_t ctxt[sizeof(update_txt_t) - LEN_DATE - LEN_SEED - LEN_PK];
  uint8_t succ[LEN_PK];
};

struct update_t {
  union {
    update_txt_t txt;
    update_bxi_t bxi;
    update_bxo_t bxo;
  } u;
};

typedef struct nonce_txt_t nonce_txt_t;
typedef struct nonce_box_t nonce_box_t;
typedef struct nonce_t nonce_t;

struct nonce_txt_t {
  uint8_t date[LEN_DATE];
  uint8_t seed[LEN_SEED];
  uint8_t zero[LEN_NPAD - 1];
  uint8_t flag;
};

struct nonce_box_t {
  uint8_t raw[LEN_NONCE];
};

struct nonce_t {
  union {
    nonce_txt_t txt;
    nonce_box_t box;
  } u;
};

#endif
