
#include <errno.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <fcntl.h>
#include "file.h"
#include "zeit.h"
#include "ee.h"
#include "tweetnacl.h"
#include "devurandom.h"
#include "customization.h"

/* All the keys are here. */
key_store_t *keys;


int key_store_create(char *name)
{
  key_store_t *kb;

  kb = (key_store_t *)malloc(sizeof(key_store_t));

  if (NULL == kb) {
    return -1;
  }
  memset(kb, 0, sizeof(key_store_t));
  kb->arch = 0x0123456789abcdef;
  kb->version = EE_VERSION_KEYSTORE;
  
  if (file_savesync(name, (uint8_t *)kb, sizeof(key_store_t)) < 0) {
    free(kb);
    return -1;
  }
  return 0;
}


int key_store_map(char *name, key_store_t **kb)
{
  struct stat buf;
  int err;
  int fd;

  fd = open(name, O_RDWR|O_NOFOLLOW);

  if (fd < 0) {
    /*
     * The return code -1 signals that the failure reason might be that the
     * keystore file wasn't found because it doesn't exist yet.  In that
     * case, we want to create a keystore file for the user.
     */
    return -1;
  }
  /* From here on, we know that opening the file did work and hence it must
   * already exist.  The error must be caused by something else.
   */
  err = -2;

  if (fstat(fd, &buf) < 0) {
    goto fail;
  }
  *kb = mmap(
    NULL, sizeof(key_store_t), PROT_READ|PROT_WRITE, MAP_FILE|MAP_SHARED, fd, (off_t)0
  );
  if (MAP_FAILED == *kb) {
    goto fail;
  }
  /* We are good to go.
   */
  err = 0;

 fail:
  close(fd);
  return err;
}


int key_store_init(char *storename)
{
  int err;

  err = key_store_map(storename, &keys);

  if ((-1 == err) && (ENOENT == errno)) {
    err = key_store_create(storename);

    if (err >= 0) {
      err = key_store_map(storename, &keys);
    }
  }
  if (0 > err) {
    return -1;
  }
  if (keys->arch != 0x0123456789abcdef) {
    fprintf(stderr, "Error, the keystore has a different architecture.\n");
    return -1;
  }
  if (keys->version != EE_VERSION_KEYSTORE) {
    fprintf(
      stderr,
      "Error, the keystore version is %u, expected is %u.\n",
      keys->version,
      EE_VERSION_KEYSTORE
    );
    return -1;
  }
  return 0;
}


int find_nick(char *nick, key_entry_t **out)
{
  key_entry_t *e;
  uint32_t i;

  for (i=0; i<keys->count; i++) {
    e = &keys->entry[i];

    if ((e->flags & FLAG_ME) == FLAG_ME) {
      if (strncmp(e->nick, nick, sizeof(e->nick)) == 0) {
	*out = e;
	return 0;
      }
    }
  }
  return NICK_UNKNOWN;
}


int reserve_unused(key_entry_t **out)
{
  if (KEYS_PER_STORE <= keys->count) {
    return STORE_OVERFLOW;
  }
  *out = &keys->entry[keys->count];
  keys->count++;
  return 0;
}


int find_sender(uint8_t a[LEN_KID], uint8_t b[LEN_KID], key_entry_t **out)
{
  key_entry_t *e;
  uint32_t i;

  for (i=0; i<keys->count; i++) {
    e = &keys->entry[i];

    if ((e->flags & (FLAG_ME | FLAG_EM)) == (FLAG_ME | FLAG_EM)) {
      if (memcmp(  e->me.pk, a, (size_t)LEN_KID) == 0) {
	if (memcmp(e->em.pk, b, (size_t)LEN_KID) == 0) {
	  *out = e;
	  return FLAG_ME;
	}
      }
      if (memcmp(  e->me.pk, b, (size_t)LEN_KID) == 0) {
	if (memcmp(e->em.pk, a, (size_t)LEN_KID) == 0) {
	  *out = e;
	  return FLAG_EM;
	}
      }
    }
  }
  return KEY_UNKNOWN;
}


int find_me(uint8_t pk[LEN_PK], key_entry_t **out)
{
  key_entry_t *e;
  uint32_t i;

  for (i=0; i<keys->count; i++) {
    e = &keys->entry[i];

    if ((e->flags & FLAG_ME) == FLAG_ME) {
      if (memcmp(e->me.pk, pk, sizeof(e->me.pk)) == 0) {
	*out = e;
	return FLAG_ME;
      }
    }
  }
  return KEY_UNKNOWN;
}


int find_em(uint8_t pk[LEN_PK], key_entry_t **out)
{
  key_entry_t *e;
  uint32_t i;

  for (i=0; i<keys->count; i++) {
    e = &keys->entry[i];

    if ((e->flags & FLAG_EM) == FLAG_EM) {
      if (memcmp(e->em.pk, pk, sizeof(e->em.pk)) == 0) {
	*out = e;
	return FLAG_EM;
      }
    }
  }
  return KEY_UNKNOWN;
}


int delete_entry(key_entry_t *e)
{
  key_entry_t *t;

  t = &keys->entry[keys->count - 1];

  if (t != e) {
    memcpy(e, t, sizeof(key_entry_t));
  }
  memset(t, 0, sizeof(key_entry_t));
  keys->count--;
  return 0;
}


int delete_nick(char *nick)
{
  key_entry_t *e;

  if (find_nick(nick, &e) < 0) {
    return NICK_UNKNOWN;
  }
  return delete_entry(e);
}


int create_nick(char *nick, char *tags, key_entry_t **out)
{
  key_entry_t *e;

  if (find_nick(nick, &e) >= 0) {
    return NICK_EXISTS;
  }
  if (reserve_unused(&e) < 0) {
    return STORE_OVERFLOW;
  }
  memset(e, 0, sizeof(key_entry_t));

  crypto_box_keypair(e->me.pk, e->me.sk);
  strncpy(e->nick, nick, sizeof(e->nick)-1);
  strncpy(e->tags, tags, sizeof(e->tags)-1);
  e->stat_date = zeit_now();
  e->flags = FLAG_ME;
  *out = e;
  return 0;
}


int set_em(key_entry_t **inout, uint8_t pk[LEN_PK])
{
  key_entry_t *e1;
  key_entry_t *e2;

  e1 = *inout;

  if ((e1->flags & FLAG_EM) == FLAG_EM) {
    return KEY_EXISTS;
  }
  if (find_me(pk, &e2) >= 0) {
    /*
     * Careful with this error.  A malicious party may have tried to send
     * Alice her own public key back.
     */
    *inout = e2;
    return KEY_ME;
  }
  if (find_sender(e1->me.pk, pk, &e2) >= 0) {
    *inout = e2;
    return KEY_IN_USE;
  }
  memcpy(e1->em.pk, pk, sizeof(e1->em.pk));
  e1->flags |= FLAG_EM;
  return 0;
}


int create_envelope(envelope_t *env, nonce_t *nonce, char *nick)
{
  key_entry_t *e;
  uint64_t now;
  int err;

  if (find_nick(nick, &e) < 0) {
    return NICK_UNKNOWN;
  }
  if ((e->flags & FLAG_EM) != FLAG_EM) {
    return KEY_UNKNOWN;
  }
  /* Prepare the nonce.
   */
  now = zeit_now();
  zeit_encode(now, nonce->u.txt.date);
  randombytes(nonce->u.txt.seed, sizeof(nonce->u.txt.seed));
  memset(nonce->u.txt.zero, 0, sizeof(nonce->u.txt.zero));

  /* Encode the direction flag so that adversaries cannot swap sender and
   * recipient.
   */
  nonce->u.txt.flag = (0 < memcmp(e->me.pk, e->em.pk, sizeof(e->me.pk))) ? 0 : 1;

  /* Prepare for encryption and encrypt.
   */
  memset(env->u.txt.date, 0, sizeof(env->u.txt.date));
  memset(env->u.txt.seed, 0, sizeof(env->u.txt.seed));
  memset(env->u.txt.zero, 0, sizeof(env->u.txt.zero));

  err = crypto_box(
    env->u.box.ctxt,
    env->u.box.ctxt,
    sizeof(env->u.box.ctxt),
    nonce->u.box.raw,
    e->em.pk,
    e->me.sk
  );

  if (0 > err) {
    return CRYPTO_ERROR;
  }
  /* Copy remaining data into place.
   */
  memcpy(env->u.txt.id_send, e->me.pk, sizeof(env->u.txt.id_send));
  memcpy(env->u.txt.id_recv, e->em.pk, sizeof(env->u.txt.id_recv));
  memcpy(env->u.txt.date, nonce->u.txt.date, sizeof(nonce->u.txt.date));
  memcpy(env->u.txt.seed, nonce->u.txt.seed, sizeof(nonce->u.txt.seed));

  /* Keep statistics.
   */
  e->stat_enc++;
  return 0;
}


int open_envelope(envelope_t *env, nonce_t *nonce, key_entry_t **out)
{
  key_entry_t *e;
  uint8_t flag;
  int err1;
  int err2;

  /* Remember the error for later when we check for a key roll-over.
   */
  err1 = find_sender(env->u.txt.id_send, env->u.txt.id_recv, &e);

  if (0 > err1) {
    return NICK_UNKNOWN;
  }
  *out = e;

  /* Prepare the nonce.
   */
  memcpy(nonce->u.txt.date, env->u.txt.date, sizeof(nonce->u.txt.date));
  memcpy(nonce->u.txt.seed, env->u.txt.seed, sizeof(nonce->u.txt.seed));
  memset(nonce->u.txt.zero, 0, sizeof(nonce->u.txt.zero));

  /* Fill in the direction flag
   */
  flag = (0 < memcmp(e->me.pk, e->em.pk, sizeof(e->me.pk))) ? 0 : 1;

  if (FLAG_EM == err1) {
    flag = 1 - flag;
  }
  nonce->u.txt.flag = flag;

  /* Prepare for encryption and encrypt.
   */
  memset(env->u.txt.date, 0, sizeof(env->u.txt.date));
  memset(env->u.txt.seed, 0, sizeof(env->u.txt.seed));

  err2 = crypto_box_open(
    env->u.box.ctxt,
    env->u.box.ctxt,
    sizeof(env->u.box.ctxt),
    nonce->u.box.raw,
    e->em.pk,
    e->me.sk
  );

  /* Copy remaining data back into place.
   */
  memcpy(env->u.txt.date, nonce->u.txt.date, sizeof(env->u.txt.date));
  memcpy(env->u.txt.seed, nonce->u.txt.seed, sizeof(env->u.txt.seed));

  if (0 > err2) {
    return CRYPTO_ERROR;
  }
  /* Keep statistics.
   */
  e->stat_dec++;
  return err1;
}


/* We establish a link from entry a -> entry b directed at entry a.
 */
int create_update(update_t *upd, nonce_t *nonce, key_entry_t *a, key_entry_t *b)
{
  uint64_t now;
  int err;

  if ((a->flags & FLAG_EM) != FLAG_EM) {
    return KEY_UNKNOWN;
  }
  /* Prepare the nonce.
   */
  now = zeit_now();
  zeit_encode(now, nonce->u.txt.date);
  randombytes(nonce->u.txt.seed, sizeof(nonce->u.txt.seed));
  memset(nonce->u.txt.zero, 0, sizeof(nonce->u.txt.zero));

  /* Prepare for encryption and encrypt.
   */
  memset(upd->u.txt.date, 0, sizeof(upd->u.txt.date));
  memset(upd->u.txt.seed, 0, sizeof(upd->u.txt.seed));
  memset(upd->u.txt.zero, 0, sizeof(upd->u.txt.zero));

  memcpy(upd->u.txt.pred, a->me.pk, sizeof(upd->u.txt.pred));
  memcpy(upd->u.txt.succ, b->me.pk, sizeof(upd->u.txt.succ));

  err = crypto_box(
    upd->u.bxi.ctxt,
    upd->u.bxi.ctxt,
    sizeof(upd->u.bxi.ctxt),
    nonce->u.box.raw,
    a->em.pk,
    b->me.sk
  );

  if (0 > err) {
    return CRYPTO_ERROR;
  }
  err = crypto_box(
    upd->u.bxo.ctxt,
    upd->u.bxo.ctxt,
    sizeof(upd->u.bxo.ctxt),
    nonce->u.box.raw,
    a->em.pk,
    a->me.sk
  );

  if (0 > err) {
    return CRYPTO_ERROR;
  }
  /* Copy remaining data into place.
   */
  memcpy(upd->u.txt.date, nonce->u.txt.date, sizeof(nonce->u.txt.date));
  memcpy(upd->u.txt.seed, nonce->u.txt.seed, sizeof(nonce->u.txt.seed));
  return 0;
}


int open_update(update_t *upd, nonce_t *nonce, key_entry_t *a, key_entry_t *b)
{
  int err;

  if ((a->flags & FLAG_EM) != FLAG_EM) {
    return KEY_UNKNOWN;
  }
  /* Prepare the nonce.
   */
  memcpy(nonce->u.txt.date, upd->u.txt.date, sizeof(nonce->u.txt.date));
  memcpy(nonce->u.txt.seed, upd->u.txt.seed, sizeof(nonce->u.txt.seed));
  memset(nonce->u.txt.zero, 0, sizeof(nonce->u.txt.zero));

  /* Prepare for decryption and decrypt.
   */
  memset(upd->u.txt.date, 0, sizeof(upd->u.txt.date));
  memset(upd->u.txt.seed, 0, sizeof(upd->u.txt.seed));

  err = crypto_box_open(
    upd->u.bxo.ctxt,
    upd->u.bxo.ctxt,
    sizeof(upd->u.bxo.ctxt),
    nonce->u.box.raw,
    a->em.pk,
    a->me.sk
  );

  /* Copy remaining data back into place.
   */
  memcpy(upd->u.txt.date, nonce->u.txt.date, sizeof(upd->u.txt.date));
  memcpy(upd->u.txt.seed, nonce->u.txt.seed, sizeof(upd->u.txt.seed));

  if (0 > err) {
    return CRYPTO_ERROR;
  }
  err = crypto_box_open(
    upd->u.bxi.ctxt,
    upd->u.bxi.ctxt,
    sizeof(upd->u.bxi.ctxt),
    nonce->u.box.raw,
    upd->u.txt.succ,
    a->me.sk
  );

  if (0 > err) {
    return CRYPTO_ERROR;
  }
  if (memcmp(a->em.pk, upd->u.txt.pred, sizeof(a->em.pk)) != 0) {
    return CRYPTO_ERROR;
  }
  /* Set up or verify the new key.
   */
  if ((b->flags & FLAG_EM) == FLAG_EM) {
    if (memcmp(b->em.pk, upd->u.txt.succ, sizeof(b->em.pk)) != 0) {
      return CRYPTO_ERROR;
    }
  }
  else {
    memcpy(b->em.pk, upd->u.txt.succ, sizeof(b->em.pk));
    b->flags |= FLAG_EM;
  }
  /* Make checks carry over from the old to the new nick.  We remove the
   * check mark on the old nick to emphasize that the new nick should be
   * used and old nicks should not be trusted any longer because their keys
   * have technically expired.
   */
  if ((a->flags & FLAG_CHECKED) == FLAG_CHECKED) {
    b->flags |= FLAG_CHECKED;
    a->flags &= (~FLAG_CHECKED);
  }
  return 0;
}


int create_efile_helper(size_t len, int max, size_t *clen, size_t *h, size_t *b, size_t *d)
{
  size_t num;
  size_t hi;
  size_t bi;
  size_t di;

  if (0 > max) {
    return -1;
  }
  num = (size_t)max;
  hi = 2;
  bi = hi + num * sizeof(envelope_t);
  di = bi + LEN_MZERO;

  *clen = di + len;
  *h = hi;
  *b = bi;
  *d = di;
  return 0;
}


int create_efile(uint8_t *c, size_t clen, char **nick, uint32_t max, size_t h, size_t b)
{
  envelope_t *env;
  uint32_t i;
  uint8_t key[LEN_KEY];
  nonce_t nonce;
  int res;
  int err;

  c[0] = (uint8_t)(max >> 8);
  c[1] = (uint8_t)(max & 0xff);
  env  = (envelope_t *)&c[h];
  res  = 0;

  randombytes(key, LEN_KEY);

  for (i=0; i<max; i++) {
    memmove(env[i].u.txt.key, key, LEN_KEY);

    err = create_envelope(&env[i], &nonce, nick[i]);

    if (NICK_UNKNOWN == err) {
      fprintf(stderr, "Error: Nick %s is unknown.\n", nick[i]);
      res = -1;
      continue;
    }
    if (KEY_UNKNOWN == err) {
      fprintf(stderr, "Error: Please ask nick %s to send his key.\n", nick[i]);
      res = -1;
      continue;
    }
    if (0 > err) {
      fprintf(stderr, "Error: Crypto failed for nick %s, ask a wizard.\n", nick[i]);
      res = -1;
      break;
    }
  }
  if (0 > res) {
    goto fail;
  }
  /* Since we use a randomly generated data encryption key for each message,
   * we set the nonce to a constant.  This is secure (according to theory)
   * and the easiest approach.  It also gives us the most room for
   * expansion.  We can add message forwarding easily, for example.  The
   * downside is that adversaries can remove recipients without being
   * detected.  But then again, they can just drop messages as well.
   */
  memset(nonce.u.box.raw, 0, sizeof(nonce.u.box.raw));
  memset(&c[b], 0, LEN_MZERO);
  crypto_secretbox(&c[b], &c[b], clen - b, nonce.u.box.raw, key);

 fail:
  memset(key, 0, LEN_KEY);
  return res;
}


/* Pre: c and clen are the address and length of the buffer of the encrypted
 * file.  Post: max is the number of envelopes, c[h] is the beginning of the
 * envelopes, c[b] is the beginning of the body.
 */
int open_efile_helper(uint8_t *c, size_t clen, uint32_t *max, size_t *h, size_t *b)
{
  uint32_t num;
  size_t hi;
  size_t bi;
  size_t di;
  size_t i;
  size_t j;

  num = (c[0] << 8) | c[1];
  
  if (clen / sizeof(envelope_t) < num) {
    return BAD_FORMAT;
  }
  hi = 2;
  bi = hi + num * sizeof(envelope_t);
  di = bi + LEN_MZERO;

  if ((di > clen) || (di < bi)) {
    return BAD_FORMAT;
  }
  for (i=0, j=bi; i<LEN_BZERO; i++, j++) {
    if (0 != c[j]) {
      return BAD_FORMAT;
    }
  }
  *max = num;
  *h = hi;
  *b = bi;
  return 0;
}


int open_efile(uint8_t *c, size_t clen, key_entry_t **pe, size_t *offs)
{
  envelope_t *env;
  key_entry_t *e;
  uint64_t date;
  uint32_t max;
  uint32_t i;
  nonce_t nonce;
  size_t bi;
  size_t hi;
  char *fmt;
  char *msg;
  char tmp[LEN_ZEIT];
  int yes;
  int err;

  err = open_efile_helper(c, clen, &max, &hi, &bi);

  if (0 > err) {
    return err;
  }
  env = (envelope_t *)&c[hi];
  yes = -1;

  for (i=0; i<max; i++) {
    err = open_envelope(&env[i], &nonce, &e);

    if (0 > err) {
      continue;
    }
    *pe = e;
    yes = i;
    fmt = (e->flags & FLAG_CHECKED) ? EES_CNICK : EES_UNICK;
    msg = (FLAG_ME == err)          ? "To"      : "From";

    printf(fmt, msg, e->nick);
  }
  if (0 > yes) {
    printf("Warning: I am not a sender or receiver of this message.\n");
    return -1;
  }
  date = zeit_decode(env[yes].u.txt.date);
  (void)zeit_ctime_r(date, tmp);

  printf("Date: %s\n", tmp);

  /* Since we use a randomly generated data encryption key for each message,
   * we set the nonce to a constant.  This is secure (according to theory)
   * and the easiest approach.  It also gives us the most room for
   * expansion.  We can add message forwarding easily, for example.
   */
  memset(nonce.u.box.raw, 0, sizeof(nonce.u.box.raw));

  err = crypto_secretbox_open(
    (uint8_t *)&c[bi],
    (uint8_t *)&c[bi],
    clen - bi,
    nonce.u.box.raw,
    env[yes].u.txt.key
  );

  /* Get rid of any traces of keys in memory.  For simplicity, we delete all
   * envelopes.
   */
  memset(c, 0, bi);

  if (0 > err) {
    printf("Abort: The message body is corrupted.\n");
    return -1;
  }
  *offs = bi + LEN_MZERO;
  return err;
}


int key_store_match(key_store_t *src, size_t len)
{
  if (sizeof(key_store_t) != len) {
    fprintf(stderr, "Key store length is wrong.\n");
    return -1;
  }
  if (KEYS_PER_STORE <= src->count) {
    fprintf(stderr, "Key store is invalid.\n");
    return -1;
  }
  if (src->arch != keys->arch) {
    fprintf(stderr, "Key store architecture mismatch.\n");
    return -1;
  }
  if (src->version != keys->version) {
    fprintf(stderr, "Key store version mismatch.\n");
    return -1;
  }
  return 0;
}


#define MATCH_NICK   1
#define MATCH_PAIR   2
#define MATCH_MIRROR 4

int key_store_merge(key_store_t *src)
{
  key_entry_t *m_nick;  /* nick match */
  key_entry_t *m_pair;  /* pair match */
  key_entry_t *b;       /* backup */
  key_entry_t *e;       /* local  */
  uint32_t i;
  int err;
  int m;

  for (i=0; i < src->count; i++) {
    b = &src->entry[i];
    m = 0;

    if (0 <= find_nick(b->nick, &m_nick)) {
      m |= MATCH_NICK;
    }
    err = find_sender(b->me.pk, b->em.pk, &m_pair);

    if (0 <= err) {
      m |= (FLAG_ME == err) ? MATCH_PAIR : MATCH_MIRROR;
    }
    if (0 == m) {
      /*
       * Add the entry to our key store if there is room.
       */
      if (0 > reserve_unused(&e)) {
	break;
      }
      /* We do have room and hence we copy the entry into our key store.
       */
      memcpy(e, b, sizeof(key_entry_t));
      continue;
    }
    if (MATCH_MIRROR == (m & MATCH_MIRROR)) {
      printf("Skipping: %-20s (mirror of %s)\n", b->nick, m_pair->nick);
      continue;
    }
    if (MATCH_PAIR == m) {
      printf("Skipping: %-20s (equals %s)\n", b->nick, m_pair->nick);
      continue;
    }
    if ((MATCH_PAIR | MATCH_NICK) == m) {
      if (m_pair != m_nick) {
	printf("Skipping: %-20s (equals %s)\n", b->nick, m_pair->nick);
	continue;
      }
      if ((!flag_is_checked(m_pair)) && flag_is_checked(b)) {
	printf("Imported: %-20s (checked %s)\n", b->nick, m_pair->nick);
	m_pair->flags |= b->flags;
      }
      continue;
    }
    if (MATCH_NICK == m) {
      if (0 != memcmp(&m_nick->me, &b->me, sizeof(em_t))) {
	printf("Conflict: %-20s (public keys differ)\n", b->nick);
	continue;
      }
      if (!flag_has_em(b)) {
	continue;
      }
      if (!flag_has_em(m_nick))
      {
	printf("Imported: %-20s (public key)\n", b->nick);
	memcpy(&m_nick->em, &b->em, sizeof(em_t));
      }
      m_nick->flags |= b->flags;
      continue;
    }
  }
  if (i < src->count) {
    printf("Merged %u of %u entries.\n", i, src->count);
  }
  else {
    printf("Done with %u entries.\n", src->count);
  }
  return 0;
}
