
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "encode.h"
#include "file.h"
#include "viz.h"

/* Coordinate system: we have tiles where a line enters on one side, and
 * leaves on one, two, or three sides.  We think of these tiles as 3x3 grids
 * of which we extract the middle points of each side starting at 12 o'clock
 * moving clock-wise.  This yields 4 bits from lsb to msb that encode the
 * character we wish to see in this tile.
 *
 *      N
 *      1
 *  W 8   2 E
 *      4
 *      S
 *
 * We draw lines in a 3n x 3m matrix and then translate this representation
 * into the semigraphics characters.
 */

tile_t ascii_tile[] = {
  { 0x00, 0x20 },
  { 0x01, 0x20 }, // invalid
  { 0x02, 0x20 }, // invalid
  { 0x03, '\\' },
  { 0x04, 0x20 }, // invalid
  { 0x05, '|'  },
  { 0x06, '/'  },
  { 0x07, '+'  },
  { 0x08, 0x20 }, // invalid
  { 0x09, '/'  },
  { 0x0a, '-'  },
  { 0x0b, '+'  },
  { 0x0c, '\\' },
  { 0x0d, '+'  },
  { 0x0e, '+'  },
  { 0x0f, '+'  }
};

tile_t iso_tile[] = {
  { 0x00, 0x20 },
  { 0x01, 0x20 }, // invalid
  { 0x02, 0x20 }, // invalid
  { 0x03, 0x94 },
  { 0x04, 0x20 }, // invalid
  { 0x05, 0x82 },
  { 0x06, 0x8c },
  { 0x07, 0x9c },
  { 0x08, 0x20 }, // invalid
  { 0x09, 0x98 },
  { 0x0a, 0x80 },
  { 0x0b, 0xb4 },
  { 0x0c, 0x90 },
  { 0x0d, 0xa4 },
  { 0x0e, 0xac },
  { 0x0f, 0xbc }
};

char *latex_tile[] = {
  "\\ ",
  "!",
  "!",
  "\\textSFii{}",
  "!",
  "\\textSFxi{}",
  "\\textSFi{}",
  "\\textSFviii{}",
  "!",
  "\\textSFiv{}",
  "\\textSFx{}",
  "\\textSFvii{}",
  "\\textSFiii{}",
  "\\textSFix{}",
  "\\textSFvi{}",
  "\\textSFv{}"
};

int key_shape[] = {
  3, 1,
  8, 0,
  0, 1,
  1, 0,
  0, 1,
  17, 0,
  0, 1,
  1, 0,
  0, 3,
  -1, 0,
  0, -1,
  -1, 0,
  0, 1,
  -1, 0,
  0, -2,
  -1, 0,
  0, 2,
  -1, 0,
  0, -1,
  -1, 0,
  0, 1,
  -1, 0,
  0, -2,
  -11, 0,
  0, 1,
  -1, 0,
  0, 1,
  -8, 0,
  0, -1,
  -1, 0,
  0, -4,
  1, 0,
  0, -1
};

int key_hole[] = {
  5, 3,
  0, 2,
  4, 0,
  0, -2,
  -4, 0
};


int peek(canvas_t *canvas, int x, int y)
{
  x = x % canvas->wd;
  y = y % canvas->ht;

  if (1 == canvas->buf[y * canvas->wd + x]) {
    return 1;
  }
  return 0;
}


void poke(canvas_t *canvas, int x, int y)
{
  x = x % canvas->wd;
  y = y % canvas->ht;

  canvas->buf[y * canvas->wd + x] = 1;
}


void poke_winding(canvas_t *canvas, int x, int y, int w)
{
  if (0 == w) {
    return;
  }
  x = (3 * x) % canvas->wd;
  y = (3 * y) % canvas->ht;

  canvas->buf[y * canvas->wd + x] = (uint8_t)w;
}


int peek_winding(canvas_t *canvas, int x, int y)
{
  x = (3 * x) % canvas->wd;
  y = (3 * y) % canvas->ht;

  return (int)canvas->buf[y * canvas->wd + x];
}


/* Require: the coordinates of a mid-point.
 */
int canvas_encode(canvas_t *canvas, int x, int y)
{
  int mask;

  mask = 0;

  mask = (mask << 1) | peek(canvas, x-1, y  );
  mask = (mask << 1) | peek(canvas, x  , y+1);
  mask = (mask << 1) | peek(canvas, x+1, y  );
  mask = (mask << 1) | peek(canvas, x  , y-1);
  return mask;
}


void canvas_line(canvas_t *canvas, int x, int y, int u, int v)
{
  int sx;
  int sy;
  int w;
  int dx;
  int dy;
  int hx;
  int hy;
  int d;

  if (u > x) {
    sx = 1;
    dx = u-x;
  }
  else {
    sx = -1;
    dx = x-u;
  }
  if (v > y) {
    sy = 1;
    dy = v-y;
  }
  else {
    sy = -1;
    dy = y-v;
  }
  /* We need this in order to fill shapes with the winding rule.
   */
  w = (0 == dy) ? 0 : sy;

  hx = 3*x+1;
  hy = 3*y+1;
  poke(canvas, hx, hy);
  poke_winding(canvas, x, y, w);
  
  if (dx >= dy) {
    d  = dy + dy - dx;

    while (x != u) {
      if (d > 0) {
	x  += sx;
	hx += sx;
	poke(canvas, hx, hy);
	hx += sx;
	poke(canvas, hx, hy);
	hx += sx;
	poke(canvas, hx, hy);
	poke_winding(canvas, x, y, w);

	y  += sy;
	hy += sy;
	poke(canvas, hx, hy);
	hy += sy;
	poke(canvas, hx, hy);
	hy += sy;
	poke(canvas, hx, hy);
	poke_winding(canvas, x, y, w);

	d = d + dy + dy - dx - dx;
      }
      else {
	x  += sx;
	hx += sx;
	poke(canvas, hx, hy);
	hx += sx;
	poke(canvas, hx, hy);
	hx += sx;
	poke(canvas, hx, hy);
	poke_winding(canvas, x, y, w);

	d = d + dy + dy;
      }
    }
  }
  else {
    d = dx + dx - dy;

    while (y != v) {
      if (d > 0) {
	y  += sy;
	hy += sy;
	poke(canvas, hx, hy);
	hy += sy;
	poke(canvas, hx, hy);
	hy += sy;
	poke(canvas, hx, hy);
	poke_winding(canvas, x, y, w);

	x  += sx;
	hx += sx;
	poke(canvas, hx, hy);
	hx += sx;
	poke(canvas, hx, hy);
	hx += sx;
	poke(canvas, hx, hy);
	poke_winding(canvas, x, y, w);

	d = d + dx + dx - dy - dy;
      }
      else {
	y  += sy;
	hy += sy;
	poke(canvas, hx, hy);
	hy += sy;
	poke(canvas, hx, hy);
	hy += sy;
	poke(canvas, hx, hy);
	poke_winding(canvas, x, y, w);

	d = d + dx + dx;
      }
    }
  }
}


/* Pre: wd > 0 and ht > 0
 */
void canvas_init(canvas_t *canvas, int wd, int ht, int type)
{
  size_t uht;
  size_t uwd;

  uht = (size_t)ht;
  uwd = (size_t)wd;

  canvas->wd   = 3 * wd;
  canvas->ht   = 3 * ht;
  canvas->mod  = 3 * wd + 1;
  canvas->buf  = (uint8_t *)calloc((size_t)1, 9 * uwd * uht);
  canvas->iso  = (uint8_t *)calloc((size_t)1, (3 * uwd + 1) * uht);

  if (0 == type) {
    canvas->code = iso_tile;
  }
  else {
    canvas->code = ascii_tile;
  }
}


void canvas_free(canvas_t *canvas)
{
  free(canvas->iso);
  free(canvas->buf);
}


void canvas_fill(canvas_t *canvas, uint8_t *buf, size_t len, int mode)
{
  uint8_t *p;
  uint8_t c; /* current character */
  size_t z; /* fill counter */
  int x; /* X coord in input buffer */
  int y; /* Y coord in input buffer */
  int u; /* X coord in output buffer */
  int v; /* Y coord in output buffer */
  int w; /* code for line symbols */
  int o; /* fill mode */
  int r; /* previous contour winding code */
  int s; /* current contour winding code */

  z = 0;

  for (v=0, y=1; y < canvas->ht; y+=3, v++) {
    p = &canvas->iso[v * canvas->mod];
    o = mode;
    r = 0;

    for (u=0, x=1; x < canvas->wd; x+=3, u++) {
      w = canvas_encode(canvas, x, y);
      c = canvas->code[w].c;

      if (c != 0x20) {
	s = peek_winding(canvas, u, v);

	if ((0 != s) && (s != r)) {
	  o ^= 1;
	  r  = s;
	}
	if (c >= 0x80) {
	  *p++ = 0xe2;
	  *p++ = 0x94;
	}
	*p++ = c;
	continue;
      }
      if ((1 == o) && (z < len)) {
	c = buf[z++];
      }
      *p++ = c;
    }
    *p = 0;
  }
  //fprintf(stderr, "Filled %lu\n", z);
}


void canvas_print(canvas_t *canvas, int fd)
{
  int y;
  int v;

  for (v=0, y=1; y < canvas->ht; y+=3, v++) {
    dprintf(fd, "%s\n", &canvas->iso[v * canvas->mod]);
  }
}


/* Uncomment this code if printing LaTeX versions of keys or transfer
 * tickets is necessary.
 *
void canvas_print_latex(canvas_t *canvas, int fd)
{
  unsigned char *s;
  unsigned char c;
  int y;
  int v;
  int i;

  for (v=0, y=1; y < canvas->ht; y+=3, v++) {
    s = &canvas->iso[v * canvas->mod];

    while (*s != 0) {
      c = *s;

      switch(c) {
      case ' ':
	dprintf(fd, "\\ ");
	break;

      case '#':
	dprintf(fd, "\\#");
	break;

      case 0xe2:
	s += 2;
	c = *s;

	for (i=0; i<16; i++) {
	  if (iso_tile[i].c == c) {
	    break;
	  }
	}
	dprintf(fd, "%s", latex_tile[i]);
	break;

      default:
	dprintf(fd, "%c", c);
	break;
      }
      s++;
    }
    dprintf(fd, "\\\\\n");
  }
}
*/


void canvas_draw_box(canvas_t *canvas, int w, int h)
{
  w--;
  h--;

  canvas_line(canvas, 0, 0, 0, h);
  canvas_line(canvas, 0, 0, w, 0);
  canvas_line(canvas, w, h, 0, h);
  canvas_line(canvas, w, h, w, 0);
}


void canvas_draw_shape(canvas_t *canvas, int *shape, size_t count)
{
  size_t i;
  int x;
  int y;
  int u;
  int v;

  x = shape[0];
  y = shape[1];

  for (i=2; i<count; i+=2) {
    u = x + shape[i];
    v = y + shape[i+1];
    canvas_line(canvas, x, y, u, v);
    x = u;
    y = v;
  }
}


void key_print(int fd, uint8_t *buf, size_t len, int type, int mode)
{
  canvas_t canvas;
  int w;
  int h;

  w = 33;
  h = 9;

  canvas_init(&canvas, w, h, type);
  canvas_draw_box(&canvas, w, h);
  canvas_draw_shape(&canvas, key_shape, sizeof(key_shape)/sizeof(int));
  canvas_draw_shape(&canvas, key_hole, sizeof(key_hole)/sizeof(int));
  canvas_fill(&canvas, buf, len, mode);
  canvas_print(&canvas, fd);
  canvas_free(&canvas);
}


void update_print(int fd, uint8_t *buf, size_t len, int type, int mode)
{
  canvas_t canvas;
  int w;
  int h;
  int u;
  int v;
  int x;

  w = 33;
  h = 9;
  u = 9;
  v = 15;
  x = 7;

  canvas_init(&canvas, w, h, type);
  canvas_draw_box(&canvas, w, h);

  canvas_line(&canvas, u  , h-1    , u+x, (h-1)/2);
  canvas_line(&canvas, u+x, (h-1)/2, u  , 0      );
  canvas_line(&canvas, v  , 0      , v+x, (h-1)/2);
  canvas_line(&canvas, v+x, (h-1)/2, v  , h-1    );

  canvas_fill(&canvas, buf, len, mode);
  canvas_print(&canvas, fd);
  canvas_free(&canvas);
}

